from typing import Any, Callable, Generic, List, Tuple

from approvaltests import Options, verify_all
from approvaltests.inline.parse2 import Parse2
from approvaltests.inline.parse3 import Parse3
from approvaltests.inline.types import T1, T2, T
from approvaltests.namer.inline_comparator import InlineComparator
from approvaltests.reporters import ReporterThatAutomaticallyApproves


class Parse(Generic[T]):
    def __init__(
        self, text: str, transformer: Callable[[str], T], options: Options
    ) -> None:
        self.text = text
        self._transformer = transformer
        self.options = options

    @staticmethod
    def doc_string(*, auto_approve: bool = False) -> "Parse[str]":
        options = Options()
        if auto_approve:
            options = options.with_reporter(ReporterThatAutomaticallyApproves())

        return Parse(
            InlineComparator.get_test_method_doc_string(), lambda s: s, options=options
        )

    def get_inputs(self) -> List[T]:
        return Parse.parse_inputs(self.text, self._transformer)

    @staticmethod
    def parse_inputs(text: str, transformer: Callable[[str], T]) -> List[T]:
        lines = text.split("\n")
        lines = list(filter(lambda line: line.strip(), lines))
        inputs = [line.split("->")[0].strip() for line in lines]
        return [transformer(i) for i in inputs]

    def verify_all(self, transform: Callable[[T], Any]) -> None:
        verify_all(
            "",
            self.get_inputs(),
            lambda s: f"{s} -> {transform(s)}",
            options=self.options.inline(),
        )

    def transform(self, transform: Callable[[T], T2]) -> "Parse[T2]":
        return Parse(self.text, lambda s: transform(self._transformer(s)), self.options)

    def transform2(
        self, transform1: Callable[[str], T1], transform2: Callable[[str], T2]
    ) -> "Parse2[T1, T2]":
        def transformer(s: str) -> Tuple[T1, T2]:
            parts = s.split(",")
            parts = list(map(lambda p: p.strip(), parts))
            return (transform1(parts[0]), transform2(parts[1]))

        return Parse2(self.text, transformer, self.options)

    # intention: find "cheating" parts in this code
    # TODO: make all hard-coded stuff be more generic
    # 'int1' is not used, let's start use it
    # add in callable
    # done: verify_all signature is incorrect
    def transform3(
        self,
        transform1: Callable[[str], T1],
        transform2: Callable[[str], T2],
        int1: Callable[[str], int],
    ) -> "Parse3[T1, int, int]":
        def transformer(s: str) -> Tuple[T1, T2, int]:
            parts = s.split(",")
            parts = list(map(lambda p: p.strip(), parts))
            return (transform1(parts[0]), transform2(parts[1]), int1(parts[2]))

        return Parse3(self.text, transformer, self.options)



import argparse
from typing import Any

from typing_extensions import override

from approvaltests.core.format_wrapper import FormatWrapper


class FormatterWrapperOfArgparseNamespace(FormatWrapper):
    @override
    def wrap(self, data: Any) -> Any:
        return FormatterOfArgparseNamespace(data)

    @override
    def is_match(self, data: Any) -> bool:
        return isinstance(data, argparse.Namespace)


class FormatterOfArgparseNamespace:
    def __init__(self, result: Any):
        self.result = result

    @override
    def __str__(self) -> str:
        from approval_utilities.utils import to_json

        return to_json(vars(self.result))



import platform

from typing_extensions import override

from approvaltests.reporters.first_working_reporter import FirstWorkingReporter
from approvaltests.reporters.generic_diff_reporter import GenericDiffReporter
from approvaltests.reporters.generic_diff_reporter_config import (
    GenericDiffReporterConfig,
)


class ReportWithDiffMergeMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/DiffMerge.app/Contents/MacOS/DiffMerge",
                extra_args=["--nosplash", "%s", "%s"],
            )
        )


class ReportWithFileMergeMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/Xcode.app/Contents/Applications/FileMerge.app/Contents/MacOS/FileMerge",
                extra_args=["--nosplash", "-left", "%s", "-right", "%s"],
            )
        )


class ReportWithBeyondCompareMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/Beyond Compare.app/Contents/MacOS/bcomp",
                extra_args=[],
            )
        )


class ReportWithKaleidoscopeMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/Kaleidoscope.app/Contents/MacOS/ksdiff",
                extra_args=[],
            )
        )


class ReportWithKaleidoscope3Mac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/usr/local/bin/ksdiff",
                extra_args=[],
            )
        )


class ReportWithKdiff3Mac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/kdiff3.app/Contents/MacOS/kdiff3",
                extra_args=["%s", "%s", "-m"],
            )
        )


class ReportWithP4mergeMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/p4merge.app/Contents/MacOS/p4merge",
                extra_args=[],
            )
        )


class ReportWithTkDiffMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/TkDiff.app/Contents/MacOS/tkdiff",
                extra_args=[],
            )
        )


class ReportWithVisualStudioCodeMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/Visual Studio Code.app/Contents/Resources/app/bin/code",
                extra_args=["-d", "%s", "%s"],
            )
        )


class ReportWithAraxisMergeMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/Araxis Merge.app/Contents/Utilities/compare",
                extra_args=[],
            )
        )


class ReportWithBeyondCompare3Windows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}Beyond Compare 3/BCompare.exe",
                extra_args=[],
            )
        )


class ReportWithBeyondCompare4Windows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}Beyond Compare 4/BCompare.exe",
                extra_args=[],
            )
        )


class ReportWithBeyondCompare5Windows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}Beyond Compare 5/BCompare.exe",
                extra_args=[],
            )
        )


class ReportWithTortoiseImageDiffWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}TortoiseSVN/bin/TortoiseIDiff.exe",
                extra_args=["/left:%s", "/right:%s"],
            )
        )


class ReportWithTortoiseTextDiffWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}TortoiseSVN/bin/TortoiseMerge.exe",
                extra_args=[],
            )
        )


class ReportWithTortoiseGitImageDiffWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}TortoiseGIT/bin/TortoiseGitIDiff.exe",
                extra_args=["/left:%s", "/right:%s"],
            )
        )


class ReportWithTortoiseGitTextDiffWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}TortoiseGIT/bin/TortoiseGitMerge.exe",
                extra_args=[],
            )
        )


class ReportWithWinMergeReporterWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}WinMerge/WinMergeU.exe",
                extra_args=[],
            )
        )


class ReportWithAraxisMergeWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}Araxis/Araxis Merge/Compare.exe",
                extra_args=[],
            )
        )


class ReportWithCodeCompareWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}Devart/Code Compare/CodeCompare.exe",
                extra_args=[],
            )
        )


class ReportWithKdiff3Windows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}KDiff3/kdiff3.exe",
                extra_args=[],
            )
        )


class ReportWithVisualStudioCodeWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}Microsoft VS Code/Code.exe",
                extra_args=["-d", "%s", "%s"],
            )
        )


class ReportWithDiffMergeLinux(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/usr/bin/diffmerge",
                extra_args=["--nosplash", "%s", "%s"],
            )
        )


class ReportWithMeldMergeLinux(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/usr/bin/meld",
                extra_args=["%s", "%s"],
            )
        )


class ReportWithKdiff3Linux(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/usr/bin/kdiff3",
                extra_args=["%s", "%s", "-m"],
            )
        )


class ReportWithSublimeMergeMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/Sublime Merge.app/Contents/SharedSupport/bin/smerge",
                extra_args=["mergetool", "%s", "%s"],
            )
        )


class ReportWithSublimeMergeWindows(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="{ProgramFiles}Sublime Merge/smerge.exe",
                extra_args=["mergetool", "%s", "%s"],
            )
        )


class ReportWithSublimeMergeLinux(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/usr/bin/smerge",
                extra_args=["mergetool", "%s", "%s"],
            )
        )


class ReportWithCursorMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/Applications/Cursor.app/Contents/Resources/app/bin/code",
                extra_args=["-d", "%s", "%s"],
            )
        )


class ReportWithDiffCommandLineLinux(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/usr/bin/diff",
                extra_args=["-u", "%s", "%s"],
            )
        )


class ReportWithDiffCommandLineMac(GenericDiffReporter):
    def __init__(self) -> None:
        super().__init__(
            config=GenericDiffReporterConfig(
                name=self.__class__.__name__,
                path="/usr/bin/diff",
                extra_args=["-u", "%s", "%s"],
            )
        )


class ReportWithDiffToolOnMac(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithDiffMergeMac(),
            ReportWithFileMergeMac(),
            ReportWithBeyondCompareMac(),
            ReportWithKaleidoscopeMac(),
            ReportWithKaleidoscope3Mac(),
            ReportWithKdiff3Mac(),
            ReportWithP4mergeMac(),
            ReportWithTkDiffMac(),
            ReportWithVisualStudioCodeMac(),
            ReportWithAraxisMergeMac(),
            ReportWithSublimeMergeMac(),
            ReportWithCursorMac(),
            ReportWithDiffCommandLineMac(),
        )

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        if platform.system() != "Darwin":
            return False
        return super().report(received_path, approved_path)


class ReportWithDiffToolOnWindows(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithBeyondCompare3Windows(),
            ReportWithBeyondCompare4Windows(),
            ReportWithBeyondCompare5Windows(),
            ReportWithTortoiseImageDiffWindows(),
            ReportWithTortoiseTextDiffWindows(),
            ReportWithTortoiseGitImageDiffWindows(),
            ReportWithTortoiseGitTextDiffWindows(),
            ReportWithWinMergeReporterWindows(),
            ReportWithAraxisMergeWindows(),
            ReportWithCodeCompareWindows(),
            ReportWithKdiff3Windows(),
            ReportWithVisualStudioCodeWindows(),
            ReportWithSublimeMergeWindows(),
        )

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        if platform.system() != "Windows":
            return False
        return super().report(received_path, approved_path)


class ReportWithDiffToolOnLinux(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithDiffMergeLinux(),
            ReportWithMeldMergeLinux(),
            ReportWithKdiff3Linux(),
            ReportWithSublimeMergeLinux(),
            ReportWithDiffCommandLineLinux(),
        )

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        if platform.system() != "Linux":
            return False
        return super().report(received_path, approved_path)


class ReportWithDiffMerge(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithDiffMergeMac(),
            ReportWithDiffMergeLinux(),
        )


class ReportWithBeyondCompare(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithBeyondCompareMac(),
            ReportWithBeyondCompare3Windows(),
            ReportWithBeyondCompare4Windows(),
            ReportWithBeyondCompare5Windows(),
        )


class ReportWithKaleidoscope(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithKaleidoscopeMac(),
            ReportWithKaleidoscope3Mac(),
        )


class ReportWithKdiff3(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithKdiff3Mac(),
            ReportWithKdiff3Windows(),
            ReportWithKdiff3Linux(),
        )


class ReportWithVisualStudioCode(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithVisualStudioCodeMac(),
            ReportWithVisualStudioCodeWindows(),
        )


class ReportWithAraxisMerge(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithAraxisMergeMac(),
            ReportWithAraxisMergeWindows(),
        )


class ReportWithTortoise(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithTortoiseImageDiffWindows(),
            ReportWithTortoiseTextDiffWindows(),
        )


class ReportWithTortoiseGit(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithTortoiseGitImageDiffWindows(),
            ReportWithTortoiseGitTextDiffWindows(),
        )


class ReportWithSublimeMerge(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithSublimeMergeMac(),
            ReportWithSublimeMergeWindows(),
            ReportWithSublimeMergeLinux(),
        )


class ReportWithDiffCommandLine(FirstWorkingReporter):
    def __init__(self) -> None:
        super().__init__(
            ReportWithDiffCommandLineLinux(),
            ReportWithDiffCommandLineMac(),
        )



class PytestConfig:
    test_naming_patterns = ["test_*"]

    @staticmethod
    def set_config(config) -> None:
        PytestConfig.test_naming_patterns = config.getini("python_functions")


def set_pytest_config(config) -> None:
    PytestConfig.set_config(config)



from typing_extensions import override

from approvaltests.core.reporter import Reporter


class FirstWorkingReporter(Reporter):
    """
    A composite reporter that goes through a list
    of reporters, running the first one that is
    working on the current machine.

    This is mostly an implementation detail of other
    classes in the library, but it may be useful in scenarios
    where a team wants to supply a list of custom reporter,
    and have the first working one of these be used.

    See also MultiReporter.
    """

    def __init__(self, *reporters: Reporter) -> None:
        self.reporters = reporters

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        for reporter in self.reporters:
            try:
                success = reporter.report(received_path, approved_path)
                if success:
                    return True
            except:  # pylint: disable=bare-except
                pass

        return False

    @override
    def __str__(self):
        reporters_string = ", ".join(str(s) for s in self.reporters)
        return f"FirstWorkingReporter({reporters_string})"

    __repr__ = __str__

    @override
    def __eq__(self, other) -> bool:
        return repr(self) == repr(other)



import shutil
import unittest

from approvaltests import Options, approvals, get_default_namer, verify, verify_file
from approvaltests.file_approver import FileApprover
from approvaltests.internals.logs.approved_file_log import ApprovedFilesLog
from approvaltests.internals.logs.failed_comparison_log import FailedComparisonLog
from approvaltests.reporters.generic_diff_reporter_factory import (
    GenericDiffReporterFactory,
)
from approvaltests.reporters.report_quietly import ReportQuietly
from approvaltests.reporters.testing_reporter import ReporterForTesting
from approvaltests.string_writer import StringWriter


class FileApproverTests(unittest.TestCase):
    @staticmethod
    def test_compare_same_files() -> None:
        writer = StringWriter("a")
        writer.write_received_file("a.txt")
        shutil.copy("a.txt", "a_same.txt")
        FileApprover.verify_files("a.txt", "a_same.txt", None, Options().comparator)

    def test_compare_different_files(self) -> None:
        reporter = ReporterForTesting()
        FileApprover.verify_files("a.txt", "b.txt", reporter, Options().comparator)
        self.assertTrue(reporter.called)

    def test_scrubbed_files(self) -> None:
        verify_file("a.txt", options=Options().with_scrubber(lambda t: "<scrubbed>"))

    def test_full(self) -> None:
        namer = get_default_namer()
        writer = StringWriter("b")
        reporter = ReporterForTesting()
        FileApprover.verify(namer, writer, reporter, Options().comparator)
        self.assertTrue(reporter.called)

    def test_returns_error_when_files_are_different(self) -> None:
        approvals.settings().allow_multiple_verify_calls_for_this_method()
        namer = get_default_namer()
        writer = StringWriter("b")
        reporter = ReporterForTesting()
        error = FileApprover.verify(namer, writer, reporter, Options().comparator)
        import re

        replaced = re.sub("ved: .*approved_files.", "ved: <rootdir>/", error)

        verify(replaced)

    def test_returns_none_when_files_are_same_files(self) -> None:
        namer = get_default_namer()
        writer = StringWriter("b")
        reporter = GenericDiffReporterFactory().get_first_working()
        error = FileApprover.verify(namer, writer, reporter, Options().comparator)
        self.assertEqual(None, error)

    def test_approved_file_is_logged(self) -> None:
        name = approvals.get_default_namer().get_approved_filename()
        name1 = name.replace(".txt", ".txt1")
        name2 = name.replace(".txt", ".txt2")

        log = ApprovedFilesLog.get_approved_files_log()
        log_lines = log.read_text().split("\n")

        # check log is cleared
        self.assertNotIn(name1, log_lines)
        self.assertNotIn(name2, log_lines)

        # touch approved file
        verify("a", options=Options().for_file.with_extension(".txt1"))
        verify("a", options=Options().for_file.with_extension(".txt2"))

        # assert that the approved file is logged
        log_lines = log.read_text().split("\n")
        self.assertIn(name1, log_lines)
        self.assertIn(name2, log_lines)

    def test_failed_comparison_is_logged(self) -> None:
        approved_name = approvals.get_default_namer().get_approved_filename()
        received_name = approvals.get_default_namer().get_received_filename()
        expected_line = f"{received_name} -> {approved_name}"

        name1 = expected_line.replace(".txt", ".txt1")
        name2 = expected_line.replace(".txt", ".txt2")

        log = FailedComparisonLog.get_failed_comparison_log()

        log_lines = log.read_text().split("\n")

        # check log is cleared
        self.assertNotIn(name1, log_lines)
        self.assertNotIn(name2, log_lines)

        # fail a verify and log it
        self.run_a_failing_test(".txt1")
        self.run_a_failing_test(".txt2")

        # assert that the approved file is logged
        log_lines = log.read_text().split("\n")
        self.assertIn(name1, log_lines)
        self.assertIn(name2, log_lines)

    def run_a_failing_test(self, extension: str) -> None:
        try:
            verify(
                "a",
                options=Options()
                .for_file.with_extension(extension)
                .with_reporter(ReportQuietly()),
            )
            self.fail("expected to fail")
        except:
            pass



from pathlib import Path
from textwrap import dedent
from typing import List

from typing_extensions import override

from approval_utilities.utilities.os_utilities import run_command
from approval_utilities.utils import ensure_file_exists, filter_values, to_json
from approvaltests.command import Command
from approvaltests.core.reporter import Reporter
from approvaltests.reporters.generic_diff_reporter_config import (
    GenericDiffReporterConfig,
    create_config,
)

PROGRAM_FILES = "{ProgramFiles}"


class GenericDiffReporter(Reporter):
    """
    A reporter that launches
    an external diff tool given by config.
    """

    limit_count = 0
    diff_tool_limit = 5
    opened_diff_tool_count = 0

    @staticmethod
    def create(diff_tool_path: str) -> "GenericDiffReporter":
        return GenericDiffReporter(create_config(["custom", diff_tool_path]))

    def __init__(self, config: GenericDiffReporterConfig) -> None:
        self.name = config.name
        self.path = self.expand_program_files(config.path)
        self.extra_args = config.extra_args

    @override
    def __str__(self) -> str:
        config = filter_values(
            lambda v: bool(v),
            {"name": self.name, "path": self.path, "arguments": self.extra_args},
        )
        return to_json(config)

    @staticmethod
    def run_command(command_array: List[str]) -> None:
        run_command(command_array)

    def get_command(self, received: str, approved: str) -> List[str]:
        if any("%s" in arg for arg in self.extra_args):
            args = []
            substitutions = [received, approved]
            sub_index = 0
            for arg in self.extra_args:
                if "%s" in arg and sub_index < len(substitutions):
                    args.append(arg.replace("%s", substitutions[sub_index], 1))
                    sub_index += 1
                else:
                    args.append(arg)
            return [self.path] + args
        return [self.path] + self.extra_args + [received, approved]

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        if not self.is_working():
            return False

        GenericDiffReporter.opened_diff_tool_count += 1

        if (
            GenericDiffReporter.diff_tool_limit
            < GenericDiffReporter.opened_diff_tool_count
        ):
            GenericDiffReporter.limit_count += 1
            print(
                dedent(f"""\
                    Skipping the diff because the limit[{GenericDiffReporter.diff_tool_limit}]
                    of opened diff reporters has been exceeded.
                    You can configure this with GenericDiffReporter.diff_tool_limit""")
            )
            return True

        ensure_file_exists(approved_path)
        command_array = self.get_command(received_path, approved_path)
        self.run_command(command_array)
        return True

    def is_working(self) -> bool:
        found = Command(self.path).locate()
        if found:
            self.path = found
            return True
        return False

    @staticmethod
    def get_limit_count() -> int:
        return GenericDiffReporter.limit_count

    @staticmethod
    def expand_program_files(path: str) -> str:
        def normalize_path(p: str) -> str:
            return Path(p).as_posix()

        if PROGRAM_FILES not in path:
            return path

        for candidate in [
            r"C:/Program Files/",
            r"C:/Program Files (x86)/",
            r"C:/ProgramW6432/",
        ]:
            possible = path.replace(PROGRAM_FILES, candidate)
            if Command.executable(possible):
                return normalize_path(possible)
        return normalize_path(path.replace(PROGRAM_FILES, "C:/Program Files/"))

    @staticmethod
    def reset_opened_diff_tool_count() -> None:
        GenericDiffReporter.opened_diff_tool_count = 0
        GenericDiffReporter.limit_count = 0



import os
from typing import Callable

from typing_extensions import override

from approval_utilities.utilities.os_utilities import run_command
from approvaltests import Reporter


def is_git_registration_needed_for_github() -> bool:
    return os.environ.get("GITHUB_ACTIONS") is not None


def run(*args: str) -> None:
    run_command(list(args))


class FileCaptureReporter(Reporter):
    def __init__(
        self,
        message: str = "*** adding received file via FileCaptureReporter for further inspection",
        is_git_registration_needed: Callable[
            [], bool
        ] = is_git_registration_needed_for_github,
    ):
        self.message = message
        self.is_git_registration_needed = is_git_registration_needed

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        if self.is_git_registration_needed():
            run("git", "config", "--local", "user.email", "action@github.com")
            run("git", "config", "--local", "user.name", "githubAction")
        run("git", "add", "--force", received_path)
        run("git", "commit", "-m", "'" + self.message + "'")
        run("git", "push")
        return True



from approvaltests import ApprovalException, Options, verify
from approvaltests.reporter_missing_exception import ReporterMissingException


def test_ApprovalException_repr() -> None:
    """
    ApprovalException('hello')
    """
    verify(repr(ApprovalException("hello")), options=Options().inline())


def test_ReporterMissingException_repr() -> None:
    """
    ReporterMissingException('MY_REPORTER')
    """
    verify(repr(ReporterMissingException("MY_REPORTER")), options=Options().inline())


def test_ReporterMissingException_str() -> None:
    """
    Could not find 'MY_REPORTER' in the environment, perhaps you need to configure your reporter.
    """
    verify(str(ReporterMissingException("MY_REPORTER")), options=Options().inline())



import os

from approvaltests.internals.logs.log_commons import (
    APPROVAL_TESTS_TEMP_DIRECTORY,
    LogCommons,
)


def test_temp_directory_is_setup_automatically() -> None:
    # setup happens automatically
    gitignore_path = APPROVAL_TESTS_TEMP_DIRECTORY / ".gitignore"
    contents = gitignore_path.read_text()
    assert contents == "*"


def test_script_download_disabled_when_env_var_set() -> None:
    os.environ["APPROVALTESTS_DISABLE_SCRIPT_DOWNLOADS"] = "1"

    result = LogCommons.download_script_from_common_repo_if_needed("test_script.py")

    assert result is False



from typing_extensions import override


class ApprovalException(Exception):
    def __init__(self, value: str) -> None:
        super().__init__(value)
        self.value = value

    @override
    def __str__(self) -> str:
        return self.value


class FrameNotFound(ApprovalException):
    pass



from typing import Any, Callable, Generic, Tuple

from approvaltests import verify_all
from approvaltests.core.options import Options
from approvaltests.inline.types import NT1, NT2, T1, T2


class Parse2(Generic[T1, T2]):
    def __init__(
        self, text: str, transformer: Callable[[str], Tuple[T1, T2]], options: Options
    ) -> None:
        self.text = text
        self._transformer = transformer
        self.options = options

    def verify_all(self, transform: Callable[[T1, T2], Any]) -> None:
        from approvaltests.inline.parse import Parse

        verify_all(
            "",
            Parse.parse_inputs(self.text, self._transformer),
            lambda s: f"{s[0]}, {s[1]} -> {transform(s[0], s[1])}",
            options=self.options.inline(),
        )

    def transform2(
        self, transform1: Callable[[T1], NT1], transform2: Callable[[T2], NT2]
    ) -> "Parse2[NT1, NT2]":
        def transformer(s: str) -> Tuple[NT1, NT2]:
            t1, t2 = self._transformer(s)
            return (transform1(t1), transform2(t2))

        return Parse2(self.text, transformer, self.options)



import re
import sys
from pathlib import Path

from approvaltests import Options, verify_argument_parser
from approvaltests.utilities.command_line_approvals import verify_command_line

_SCRIPT_DIR = Path(__file__).parent
_REPO_ROOT = _SCRIPT_DIR.parent
sys.path.insert(0, _REPO_ROOT.joinpath("internal_documentation/scripts").as_posix())
from ai_fixer_loop import (  # pylint: disable=wrong-import-position,wrong-import-order
    get_argument_parser,
)


def test_ai_fixer_loop() -> None:
    python = sys.executable
    cwd = _REPO_ROOT / "internal_documentation" / "scripts"
    verify_command_line(
        python
        + f" ai_fixer_loop.py --find test_find --fix test_fix.sh --tcr {_REPO_ROOT / 'internal_documentation/scripts/test_tcr'}",
        current_working_directory=str(cwd),
        options=Options().with_scrubber(
            lambda text: re.sub(r" \[[\d\.]+s\]", " [time]", text)
        ),
    )


def test_argument_parser() -> None:
    """
    usage: ai_fixer_loop [-h] [--find FIND] [--fix FIX] [--tcr TCR]

    AI-Powered Loop Fixer.

    <optional header>:
      -h, --help   show this help message and exit
      --find FIND  The script to run to find problems.
      --fix FIX    The script to run to fix problems.
      --tcr TCR    The script for Test && Commit || Revert.
    """
    verify_argument_parser(get_argument_parser(), options=Options().inline())



from typing import Union

from typing_extensions import override

from approval_utilities.utils import create_directory_if_needed
from approvaltests.core.writer import Writer

# typing.ByteString was removed in Python 3.14
_ByteString = Union[bytes, bytearray, memoryview]


class BinaryWriter(Writer):
    contents: _ByteString

    def __init__(
        self,
        contents: _ByteString,
        extension: str,
    ) -> None:
        self.contents = contents
        self.extension_with_dot = extension

    @override
    def write_received_file(self, received_file: str) -> str:
        create_directory_if_needed(received_file)
        with open(received_file, mode="wb") as file:
            file.write(self.contents)

        return received_file



import importlib
import inspect
import sys

from approvaltests import (
    DiffReporter,
    ReporterForTesting,
    approvals,
    combination_approvals,  # noqa: F401
    get_default_reporter,
    verify,
    verify_all,
)
from approvaltests.core.options import Options
from approvaltests.reporters import MultiReporter, ReportByCreatingDiffFile
from approvaltests.utilities import command_line_approvals  # noqa: F401
from approvaltests.utilities.logger import simple_logger_approvals  # noqa: F401
from approvaltests.utilities.logging import logging_approvals  # noqa: F401

_approvals_modules = list(
    sorted(
        filter(
            lambda name: (
                name.startswith("approvaltests.") and name.endswith("approvals")
            ),
            sys.modules.keys(),
        )
    )
)


def test_list_of_modules() -> None:
    verify_all("", _approvals_modules)


def test_every_function_in_approvals_with_verify_has_an_options() -> None:
    for module_name in _approvals_modules:
        assert_verify_methods_have_options(importlib.import_module(module_name))


def assert_verify_methods_have_options(module) -> None:
    for function_name, obj in module.__dict__.items():
        if "verify" not in function_name:
            continue

        if not callable(obj):
            continue

        # if it has a decorator, we need to get the original function
        if hasattr(obj, "__wrapped__"):
            obj = obj.__wrapped__

        argspec = inspect.getfullargspec(obj)
        has_options = "options" in argspec.kwonlyargs
        assert has_options, (
            f"Missing Keyword only parameter `options` in {function_name}:\n\t{argspec}"
        )


def test_empty_options_has_default_reporter() -> None:
    ##approvals.set_default_reporter(None)
    options = Options()
    assert options.reporter == get_default_reporter()


def test_with_reporter() -> None:
    testr = ReporterForTesting()
    options = Options().with_reporter(testr)
    try:
        verify("Data2", options=options)
    except:
        pass

    assert testr.called


def test_setting_reporter() -> None:
    testr = ReporterForTesting()
    options = Options().with_reporter(testr)
    assert options.reporter == testr


def test_file_extensions() -> None:
    approvals.settings().allow_multiple_verify_calls_for_this_method()
    content = "# This is a markdown header\n"
    # begin-snippet: options_with_file_extension
    verify(content, options=Options().for_file.with_extension(".md"))
    # end-snippet
    verify(content, options=Options().for_file.with_extension("md"))


def test_overwrite_reporter() -> None:
    # current behaviour, override
    options0 = (
        Options()
        .with_reporter(ReportByCreatingDiffFile())
        .with_reporter(DiffReporter())
    )
    assert type(options0.reporter) == DiffReporter


def test_add_reporter() -> None:
    reporter1 = ReportByCreatingDiffFile()
    reporter2 = DiffReporter()
    handmade = MultiReporter(reporter1, reporter2)

    options0 = Options().with_reporter(reporter1).add_reporter(reporter2)
    assert str(options0.reporter) == str(handmade)



import sys
import unittest

import pytest

from approvaltests import (
    ApprovalException,
    Options,
    verify,
    verify_all,
    verify_all_combinations_with_labeled_input,
)
from approvaltests.inline.inline_options import InlineOptions
from approvaltests.inline.markers import PRESERVE_LEADING_WHITESPACE_MARKER
from approvaltests.inline.parse_docstring import parse_docstring
from approvaltests.namer.inline_comparator import InlineComparator
from approvaltests.namer.inline_python_reporter import (
    detect_trailing_whitespace,
    escape_backslashes,
    escape_characters_dict,
    handle_preceeding_whitespace,
)
from approvaltests.reporters.diff_tools import ReportWithBeyondCompare
from approvaltests.reporters.report_quietly import ReportQuietly

# Todo:
# detect the actual tab
# detect if the quote type used in the docstring  (i.e. " or ')


def verify_approval_failure(text: str) -> None:
    with pytest.raises(ApprovalException):
        verify(text, options=Options().with_reporter(ReportQuietly()).inline())


def fizz_buzz(param: int) -> str:
    return_string = ""
    for i in range(1, param + 1):
        if i % 15 == 0:
            return_string += "FizzBuzz\n"
        elif i % 3 == 0:
            return_string += "Fizz\n"
        elif i % 5 == 0:
            return_string += "Buzz\n"
        else:
            return_string += str(i) + "\n"
    return return_string


def test_fizz_buzz() -> None:
    """
    1
    2
    Fizz
    4
    Buzz
    Fizz
    7
    8
    """
    verify(fizz_buzz(8), options=Options().inline())


def test_docstrings() -> None:
    """
    hello
    world
    """
    # verify_inline(greetting())
    # verify(greetting(), options=Options().inline(show_code= False))
    verify(greeting(), options=Options().inline())


def greeting() -> str:
    return "hello\nworld"


def test_docstring_parsing() -> None:
    """
    1
    2 -> 2
    Fizz
    4
    Buzz
    Fizz
    7
    8
    """
    verify_all("inputs", parse_docstring())


def test_uppercase() -> None:
    """
    a -> A
    b -> B
    c -> C
    """
    verify(
        "\n".join([f"{a} -> {a.upper()}" for a in parse_docstring()]),
        options=Options().inline(),
    )


options = Options().inline()


def test_when_options_is_created_outside_of_test() -> None:
    """
    hello
    world
    """
    verify(greeting(), options=options)


def test_exception_on_failure() -> None:
    """
    this string should not match
    """
    with pytest.raises(ApprovalException):
        verify(greeting(), options=Options().with_reporter(ReportQuietly()).inline())


class InlineTests(unittest.TestCase):
    def test_with_labeled_input_inline(self) -> None:
        """
        (arg1: 1, arg2: 2) => 3
        (arg1: 1, arg2: 4) => 5
        (arg1: 3, arg2: 2) => 5
        (arg1: 3, arg2: 4) => 7
        """
        verify_all_combinations_with_labeled_input(
            lambda a, b: a + b,
            arg1=(1, 3),
            arg2=(2, 4),
            options=Options().inline(),
        )


def get_preceding_whitespace() -> str:
    return "    4 whitespaces"


# fmt: off
@unittest.skipIf(sys.version_info >= (3, 13), "__doc__ removes preceding whitespace in Python 3.13+")
def test_preceding_whitespace_before_python_3_13() -> None:
    """
        4 whitespaces
    """
    verify(get_preceding_whitespace(), options=Options().inline())

def test_preceding_whitespace() -> None:
    """
    <<approvaltests:preserve-leading-whitespace>>
        4 whitespaces
    """
    verify(get_preceding_whitespace(), options=Options().inline())


def test_no_preceeding_whitespace() -> None:
    """
        4 whitespaces
    applesauce
    """
    text = "    4 whitespaces\napplesauce"
    verify(text, options=Options().inline())


def test_trailing_whitespace() -> None:
    """
    4 trailing whitespaces    
    """  # Warning: Editors may remove trailing spaces, causing this test to fail

    # Note: Pycharm will remove the trailing whitespaces, to disable this go to:
    # File -> Settings -> Editor -> General -> On Save -> [ ] Remove trailing spaces
    verify("4 trailing whitespaces    ", options=Options().inline())

# fmt: on


def test_bug_blank_lines() -> None:
    """


    test bug <br \\> with blank lines



    """
    verify("\n\ntest bug <br \\> with blank lines\n\n\n\n", options=Options().inline())


def test_handle_preceeding_whitespace_all_lines_indented() -> None:
    text = "    indented with spaces\n\tindented with tab\n    still indented"
    result = handle_preceeding_whitespace(text)
    assert result.startswith(PRESERVE_LEADING_WHITESPACE_MARKER)


def test_handle_preceeding_whitespace_one_line_not_indented() -> None:
    text = "    indented with spaces\nnot indented\n\tindented with tab"
    result = handle_preceeding_whitespace(text)
    assert not result.startswith(PRESERVE_LEADING_WHITESPACE_MARKER)


def test_inline_with_additional_reporter() -> None:
    """
    hello
    world
    """
    verify(
        "hello\nworld",
        options=(Options().inline().add_reporter(ReportWithBeyondCompare())),
    )


def test_inline_with_preserved_approved_text() -> None:
    """
    42
    ***** DELETE ME TO APPROVE *****
    vvvvv PREVIOUS RESULT vvvvv
    41
    """
    options = Options().inline(InlineOptions.semi_automatic_with_previous_approved())
    try:
        verify("42", options=options)
    except ApprovalException:
        pass
    verify(InlineComparator.get_test_method_doc_string())


def test_inline_with_semi_automatic_inline() -> None:
    """
    42
    ***** DELETE ME TO APPROVE *****
    """
    options = Options().inline(InlineOptions.semi_automatic())
    try:
        verify("42", options=options)
    except ApprovalException:
        pass
    verify(InlineComparator.get_test_method_doc_string())


def test_detect_trailing_whitespace_true() -> None:
    text = "no trail\nwith trail \t\nnormal"
    assert detect_trailing_whitespace(text)


def test_detect_trailing_whitespace_false() -> None:
    text = "alpha\nbeta\ngamma"
    assert not detect_trailing_whitespace(text)


def test_detect_trailing_whitespace_ignores_pure_empty_lines() -> None:
    text = "line1\n\nline2"
    assert not detect_trailing_whitespace(text)


def test_escape_backslashes_windows_path() -> None:
    text = "C:\\Temp\\foo\\bar\\note.txt"
    assert escape_backslashes(text) == "C:\\Temp\\foo\\bar\\note.txt".replace(
        "\\", "\\\\"
    )


def test_escape_backslashes_mixed_sequences() -> None:
    text = "a\\b\\c\\n"
    assert escape_backslashes(text) == "a\\b\\c\\n".replace("\\", "\\\\")


def test_unicode_and_special_characters__passing() -> None:
    """
    é
    """
    text = "e\u0301"  # 'e' + COMBINING ACUTE (2 code points)
    verify(text, options=Options().inline())


def test_unicode_and_special_characters__mismatch() -> None:
    """
    é
    """
    # 'e' + COMBINING ACUTE (2 code points)
    verify_approval_failure("xe\u0301")


def test_unicode_and_special_characters__missing() -> None:
    text = "e\u0301"  # 'e' + COMBINING ACUTE (2 code points)
    verify_approval_failure(text)


def test_unicode_and_special_characters__not_inline() -> None:
    text = "e\u0301"  # 'e' + COMBINING ACUTE (2 code points)
    verify(text)


def test_unicode_null_character__passing() -> None:
    """
    \x00
    """
    text = "\x00"  # Null character
    verify(text, options=Options().inline())


def test_unicode_null_character__missing() -> None:
    text = "hello\x00world"  # Null character
    verify_approval_failure(text)


def test_unicode_null_character__incorrect_test() -> None:
    """
    hello\x00world
    """
    verify_approval_failure("hello world")


def test_unicode_null_character__incorrect_approval() -> None:
    """
    hello .
    """
    verify_approval_failure("hello\x00.")


def test_unicode_right_to_left_mark__passing() -> None:
    """
    hello!\u200fworld
    """
    text = "hello!\u200fworld"  # Right-to-left mark
    verify(text, options=Options().inline())


def test_unicode_right_to_left_mark__missing() -> None:
    text = "hello!\u200fworld"  # Right-to-left mark
    verify_approval_failure(text)


def test_unicode_right_to_left_mark__incorrect_approval() -> None:
    """
    hello world
    """
    verify_approval_failure("hello!\u200fworld")


def test_unicode_right_to_left_mark__incorrect_test() -> None:
    """
    hello!\u200fworld
    """
    verify_approval_failure("hello world")


def test_escape_characters_dict() -> None:
    verify_all("", escape_characters_dict.items())



import os
from typing import Dict

from _pytest.fixtures import FixtureRequest
from typing_extensions import override

from approvaltests.namer.namer_base import NamerBase


class PyTestNamer(NamerBase):
    "Pass a pytest 'request' to the constructor. You get one of those as a pytest fixture"

    def __init__(self, request: FixtureRequest, extension: None = None) -> None:
        NamerBase.__init__(self, extension)
        self.request = request
        self.filepath, self.filename = os.path.split(str(self.request.fspath))
        self.config: Dict[str, str] = {}
        subdir = self.request.config.getoption(
            "approvaltests_subdirectory", default=None
        )
        if subdir:
            self.config["subdirectory"] = subdir

    @override
    def get_file_name(self) -> str:
        return f"{os.path.splitext(self.filename)[0]}.{self.request.node.name}"

    @override
    def get_directory(self) -> str:
        return self.filepath

    @override
    def get_config(self) -> Dict[str, str]:
        return self.config



import re
from enum import Enum

from typing_extensions import override


class SplitCode:
    def __init__(self, before_method: str, after_method: str, tab: str) -> None:
        self.before_method = before_method
        self.after_method = after_method
        self.tab = tab

    @override
    def __str__(self) -> str:
        return f"before:\n{self.before_method}\nafter:\n{self.after_method}\ntab: '{self.tab}'\n"

    @staticmethod
    def on_method(code: str, method_name: str) -> "SplitCode":
        lines = code.split("\n")
        before = []
        after = []
        tab = "    "

        class State(Enum):
            BEFORE = 0
            FIRST_LINE_OF_METHOD_BODY = 1
            IN_DOCSTRING = 2
            AFTER_DOCTSTRING = 3

        state = State.BEFORE

        for line in lines:
            stripped_line = line.strip()
            if stripped_line.startswith(f"def {method_name}("):
                state = State.FIRST_LINE_OF_METHOD_BODY
                before.append(line)
            elif state == State.BEFORE:
                before.append(line)
            elif state == State.FIRST_LINE_OF_METHOD_BODY:
                tab = re.compile(r"^\s*").match(line).group()
                if stripped_line.startswith('"""'):
                    state = State.IN_DOCSTRING
                    doc_string_quotes = '"""'
                elif stripped_line.startswith("'''"):
                    state = State.IN_DOCSTRING
                    doc_string_quotes = "'''"
                else:
                    state = State.AFTER_DOCTSTRING
                    after.append(line)
            elif state == State.IN_DOCSTRING:
                if stripped_line.startswith(doc_string_quotes):
                    state = State.AFTER_DOCTSTRING
            elif state == State.AFTER_DOCTSTRING:
                after.append(line)
        return SplitCode("\n".join(before), "\n".join(after), tab)

    def indent(self, received_text: str) -> str:
        lines = received_text.split("\n")
        indented_lines = [f"{self.tab}{line}" for line in lines]
        return "\n".join(indented_lines)



from pathlib import Path

import yaml

from approvaltests import verify

SCRIPT_DIR = Path(__file__).parent.resolve()


def test_workflow_matrix_python_versions() -> None:
    test_workflow = yaml.safe_load(
        (SCRIPT_DIR.parent / ".github/workflows/test.yml").read_text()
    )
    python_versions = test_workflow["jobs"]["build"]["strategy"]["matrix"][
        "python-version"
    ]

    test_min_workflow = yaml.safe_load(
        (SCRIPT_DIR.parent / ".github/workflows/test_min.yml").read_text()
    )
    python_versions_min = test_min_workflow["jobs"]["build"]["strategy"]["matrix"][
        "python-version"
    ]
    assert python_versions == python_versions_min, (
        "Python version matrix in test.yml and test_min.yml do not match"
    )

    test_current_release_workflow = yaml.safe_load(
        (SCRIPT_DIR.parent / ".github/workflows/test_current_release.yml").read_text()
    )
    python_versions_current_release = test_current_release_workflow["jobs"]["build"][
        "strategy"
    ]["matrix"]["python-version"]
    assert python_versions == python_versions_current_release, (
        "Python version matrix in test.yml and test_current_release.yml do not match"
    )

    verify(
        f"ApprovalTests is tested on the following Python versions: {', '.join(python_versions)}."
    )



import tempfile
from inspect import FrameInfo
from pathlib import Path
from typing import Callable, Optional

from typing_extensions import override

from approvaltests import Reporter, StackFrameNamer
from approvaltests.inline.markers import PRESERVE_LEADING_WHITESPACE_MARKER
from approvaltests.inline.split_code import SplitCode


def handle_preceeding_whitespace(received_text: str) -> str:
    if not received_text:
        return received_text
    lines = received_text.split("\n")
    if all(
        (line.startswith(" ") or line.startswith("\t")) for line in lines if line != ""
    ):
        return f"{PRESERVE_LEADING_WHITESPACE_MARKER}" + received_text
    return received_text


def detect_trailing_whitespace(text: str) -> bool:
    if not text:
        return False
    return any(len(line) and line[-1] in " \t" for line in text.split("\n"))


def escape_backslashes(text: str) -> str:
    """
    Ensure literal backslashes in the approved docstring are preserved by escaping
    them in the source code injection so Python does not interpret sequences like
    "\n" or "\t".
    """
    if not text:
        return text
    return text.replace("\\", "\\\\")


escape_characters_dict = {
    v.encode().decode("unicode_escape"): v
    for v in [
        r"\x00",  # Null
        r"\b",  # Backspace
        r"\v",  # Vertical tab
        r"\f",  # Form feed
        r"\x1b",  # Escape
        r"\x7f",  # Delete
        r"\u00a0",  # Non-breaking space
        r"\u200b",  # Zero-width space
        r"\u200c",  # Zero-width non-joiner
        r"\u200d",  # Zero-width joiner
        r"\u200e",  # Left-to-right mark
        r"\u200f",  # Right-to-left mark
        r"\u202c",  # Pop directional formatting
        r"\u202d",  # Left-to-right override
        r"\u202e",  # Right-to-left override
        r"\u2028",  # Line separator
        r"\u2029",  # Paragraph separator
        r"\ufeff",  # Byte order mark
    ]
}


def escape_control_characters(text: str) -> str:
    if not text:
        return text

    translation_table = str.maketrans(escape_characters_dict)

    return text.translate(translation_table)


class InlinePythonReporter(Reporter):
    def __init__(
        self,
        reporter: Reporter,
        create_footer_function: Optional[Callable[[str, str], str]] = None,
    ):
        self.diffReporter = reporter
        self.footer_function = create_footer_function or (lambda __, ___: "")
        self.footer = ""

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        test_source_file = self.get_test_source_file()
        self.footer = self.footer_function(received_path, approved_path)
        received_path = self.create_received_file(received_path, test_source_file)
        return self.diffReporter.report(received_path, test_source_file)

    def get_test_source_file(self) -> str:
        test_stack_frame: FrameInfo = StackFrameNamer.get_test_frame()
        return test_stack_frame.filename

    def create_received_file(self, received_path: str, test_source_file: str) -> str:
        code = Path(test_source_file).read_text(encoding="utf-8")

        original_received_text = Path(received_path).read_text(encoding="utf-8")[:-1]
        received_text = original_received_text + self.footer
        received_text = handle_preceeding_whitespace(received_text)
        received_text = escape_backslashes(received_text)
        received_text = escape_control_characters(received_text)
        method_name = StackFrameNamer.get_test_frame().function
        if detect_trailing_whitespace(original_received_text):
            trailing_comment = "  # Warning: Editors may remove trailing spaces, causing this test to fail"
        else:
            trailing_comment = ""

        new_code = self.swap(
            received_text, code, method_name, after_docstring_comment=trailing_comment
        )
        file = tempfile.NamedTemporaryFile(suffix=".received.txt", delete=False).name
        Path(file).write_text(new_code, encoding="utf-8")
        return file

    def swap(
        self,
        received_text: str,
        code: str,
        method_name: str,
        after_docstring_comment: str = "",
    ) -> str:
        split_code = SplitCode.on_method(code, method_name)
        return "\n".join(
            [
                split_code.before_method,
                split_code.tab + '"""',
                split_code.indent(received_text),
                split_code.tab + '"""' + after_docstring_comment,
                split_code.after_method,
            ]
        )



from typing_extensions import override

from approvaltests.core.reporter import Reporter


class MultiReporter(Reporter):
    """
    A composite reporter that goes through a list
    of reporters, running all that are working on
    the current machine.

    See also FirstWorkingReporter.
    """

    def __init__(self, *reporters: Reporter) -> None:
        self.reporters = reporters

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        for reporter in self.reporters:
            reporter.report(received_path, approved_path)

    @override
    def __str__(self) -> str:
        return f"MultiReporter({', '.join([r.__class__.__name__ for r in self.reporters])})"



from typing import Any, Optional

from typing_extensions import override

from approval_utilities.utils import write_to_temporary_file
from approvaltests import FileApprover
from approvaltests.approvals import (
    get_default_namer,
    initialize_options,
    verify_with_namer,
)
from approvaltests.core import Reporter
from approvaltests.core.options import Options
from approvaltests.namer.stack_frame_namer import StackFrameNamer
from approvaltests.reporters.testing_reporter import ReporterForTesting


class FilePathNamer(StackFrameNamer):
    def __init__(self, file_path: str, extension: None = None) -> None:
        StackFrameNamer.__init__(self, extension)
        self.file_path = file_path

    @override
    def get_approved_filename(self, base: Optional[str] = None) -> str:
        return self.file_path


def assert_against_file(
    actual: str, file_path: str, reporter: Optional[ReporterForTesting] = None
) -> None:
    namer = FilePathNamer(file_path)
    FileApprover.add_allowed_duplicates(lambda n: n == namer.get_approved_filename())
    verify_with_namer(actual, namer, reporter)


def assert_equal_with_reporter(
    expected: str,
    actual: Any,
    reporter: Optional[Reporter] = None,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> None:
    options = initialize_options(options, reporter)
    actual = options.scrub(actual)
    if actual == expected:
        return

    name = get_default_namer().get_file_name()
    extention = options.for_file.file_extention
    expected_file = write_to_temporary_file(expected, name + ".expected.", extention)
    actual_file = write_to_temporary_file(actual, name + ".actual.", extention)
    options.reporter.report(actual_file, expected_file)
    raise AssertionError(
        f'expected != actual\n  actual: "{actual}"\nexpected: "{expected}"'
    )



from types import TracebackType
from typing import ContextManager, Optional, Type

from testfixtures import LogCapture

from approvaltests import Options, verify
from approvaltests.scrubbers.date_scrubber import DateScrubber


def verify_logging(
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    options: Optional[Options] = None,
) -> ContextManager[None]:
    class VerifyLogging:
        def __init__(self) -> None:
            self.l = LogCapture()
            self.output = "anything"
            self.options = options or Options()

        def __enter__(self) -> None:
            self.l.__enter__()

            pass

        def __exit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException],
            exc_tb: Optional[TracebackType],
        ) -> bool:
            self.l.__exit__(exc_type, exc_val, exc_tb)
            self.options = self.options.add_scrubber(
                DateScrubber.get_scrubber_for("2023-07-16 17:39:03.293919")
            )
            verify(self.l, options=self.options)

    return VerifyLogging()



from typing_extensions import override


class ReporterMissingException(BaseException):
    def __init__(self, key: str) -> None:
        super().__init__(key)
        self.value = key

    @override
    def __str__(self) -> str:
        return f"Could not find {repr(self.value)} in the environment, perhaps you need to configure your reporter."



import tempfile
from inspect import FrameInfo
from pathlib import Path
from typing import Any, Callable, Optional

from typing_extensions import override

from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from approval_utilities.utilities.stack_frame_utilities import get_class_name_for_frame
from approvaltests import Namer, StackFrameNamer
from approvaltests.core.options import Options
from approvaltests.inline.inline_options import InlineOptions
from approvaltests.inline.markers import PRESERVE_LEADING_WHITESPACE_MARKER


class InlineComparator(Namer):
    @override
    def get_approved_filename(self, base: Optional[str] = None) -> str:
        file = tempfile.NamedTemporaryFile(suffix=".approved.txt", delete=False).name
        docs = self.get_test_method_doc_string()
        Path(file).write_text(docs, encoding="utf-8")
        return file

    @override
    def get_received_filename(self, base: Optional[str] = None) -> str:
        return tempfile.NamedTemporaryFile(suffix=".received.txt", delete=False).name

    @staticmethod
    def get_test_method_doc_string() -> str:
        test_stack_frame: FrameInfo = StackFrameNamer.get_test_frame()
        method: Callable[..., Any] = InlineComparator.get_caller_method(
            test_stack_frame
        )
        doc = remove_indentation_from(method.__doc__)
        if doc.startswith(PRESERVE_LEADING_WHITESPACE_MARKER):
            return doc[len(PRESERVE_LEADING_WHITESPACE_MARKER) :]
        return doc

    @staticmethod
    def get_caller_method(caller_frame: FrameInfo) -> Callable:
        caller_function_name: str = caller_frame[3]
        caller_function_object = caller_frame.frame.f_globals.get(caller_function_name)
        if caller_function_object:
            # pytest function
            return caller_function_object
        else:
            # unittest class function
            className = get_class_name_for_frame(caller_frame)
            clazz = caller_frame.frame.f_globals.get(className)
            caller_function_object = clazz.__dict__.get(caller_function_name)
            return caller_function_object

    def register(
        self, options: Options, inline_options: Optional[InlineOptions] = None
    ) -> Options:
        inline_options = (
            InlineOptions.show_code() if inline_options is None else inline_options
        )
        options2 = options.with_namer(self)
        options2 = inline_options.apply(options2)

        return options2



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from typing_extensions import override

if TYPE_CHECKING:
    from approvaltests.core.options import Options

PREVIOUS_RESULT_ = "vvvvv PREVIOUS RESULT vvvvv\n"

DELETE_ME_TO_APPROVE_ = "\n***** DELETE ME TO APPROVE *****"


class InlineOptions:
    @staticmethod
    def automatic() -> InlineOptions:
        from approvaltests.namer.inline_python_reporter import InlinePythonReporter
        from approvaltests.reporters import ReporterThatAutomaticallyApproves

        class AutomaticInlineOptions(InlineOptions):
            @override
            def apply(self, options: Options) -> Options:
                return options.with_reporter(
                    InlinePythonReporter(ReporterThatAutomaticallyApproves())
                )

        return AutomaticInlineOptions()

    @staticmethod
    def semi_automatic() -> InlineOptions:
        from approvaltests.namer.inline_python_reporter import InlinePythonReporter
        from approvaltests.reporters import ReporterThatAutomaticallyApproves

        class SemiAutomaticInlineOptions(InlineOptions):
            @override
            def apply(self, options: Options) -> Options:
                return options.with_reporter(
                    InlinePythonReporter(
                        ReporterThatAutomaticallyApproves(),
                        create_footer_function=lambda __, ___: DELETE_ME_TO_APPROVE_,
                    )
                )

        return SemiAutomaticInlineOptions()

    @staticmethod
    def semi_automatic_with_previous_approved() -> InlineOptions:
        from approvaltests.namer.inline_python_reporter import InlinePythonReporter
        from approvaltests.reporters import ReporterThatAutomaticallyApproves

        def create_previous_capture_footer(
            received_path: str, approved_path: str
        ) -> str:
            approved_text = Path(approved_path).read_text()
            approved_text = approved_text.rsplit("\n", 1)[0]
            approved_text = approved_text.rsplit(PREVIOUS_RESULT_, 1)[-1]
            received_text = Path(received_path).read_text().rsplit("\n", 1)[0]

            if received_text != approved_text:
                return DELETE_ME_TO_APPROVE_ + "\n" + PREVIOUS_RESULT_ + approved_text
            return ""

        class PreviousCaptureInlineOptions(InlineOptions):
            @override
            def apply(self, options: Options) -> Options:
                return options.with_reporter(
                    InlinePythonReporter(
                        ReporterThatAutomaticallyApproves(),
                        create_footer_function=create_previous_capture_footer,
                    )
                )

        return PreviousCaptureInlineOptions()

    def apply(self, options: Options) -> Options:
        return options

    @staticmethod
    def show_code(do_show_code: bool = True) -> InlineOptions:
        from approvaltests.namer.inline_python_reporter import InlinePythonReporter

        class ShowCodeInlineOptions(InlineOptions):
            @override
            def apply(self, options: Options) -> Options:
                return options.with_reporter(InlinePythonReporter(options.reporter))

        class DoNotShowCodeInlineOptions(InlineOptions):
            @override
            def apply(self, options: Options) -> Options:
                return options

        return ShowCodeInlineOptions() if do_show_code else DoNotShowCodeInlineOptions()



from unittest import IsolatedAsyncioTestCase

from approvaltests import verify
from approvaltests.namer.stack_frame_namer import StackFrameNamer


class TestIsolatedAsyncioTestCase(IsolatedAsyncioTestCase):
    async def test_namer_finds_correct_method_name(self) -> None:
        namer = StackFrameNamer()
        self.assertEqual(
            "test_namer_finds_correct_method_name", namer.get_method_name()
        )

    async def test_namer_finds_correct_class_name(self) -> None:
        namer = StackFrameNamer()
        self.assertEqual(self.__class__.__name__, namer.get_class_name())

    async def test_example(self) -> None:
        verify("something")



from typing_extensions import override

from approval_utilities.utils import append_to_file, is_windows_os
from approvaltests import Reporter
from approvaltests.internals.logs.log_commons import APPROVAL_TESTS_TEMP_DIRECTORY
from approvaltests.reporters import get_command_text


class ReporterThatCreatesAnApprovalScript(Reporter):
    file = None

    def create_approval_script(self, script: str) -> None:
        if ReporterThatCreatesAnApprovalScript.file is None:
            if is_windows_os():
                self.create_script_windows()
            else:
                self.create_script_unix()
        append_to_file(ReporterThatCreatesAnApprovalScript.file, f"{script}\n")

    def create_script_unix(self) -> None:
        APPROVAL_TESTS_TEMP_DIRECTORY.mkdir(parents=True, exist_ok=True)
        ReporterThatCreatesAnApprovalScript.file = (
            APPROVAL_TESTS_TEMP_DIRECTORY / "approval_script.sh"
        )
        ReporterThatCreatesAnApprovalScript.file.write_text("#!/bin/bash\n")
        # make executable
        ReporterThatCreatesAnApprovalScript.file.chmod(0o755)

    def create_script_windows(self) -> None:
        APPROVAL_TESTS_TEMP_DIRECTORY.mkdir(parents=True, exist_ok=True)
        ReporterThatCreatesAnApprovalScript.file = (
            APPROVAL_TESTS_TEMP_DIRECTORY / "approval_script.bat"
        )
        ReporterThatCreatesAnApprovalScript.file.write_text("")

    @override
    def report(self, received_path: str, approved_path: str) -> bool:
        self.create_approval_script(get_command_text(received_path, approved_path))
        return True



from typing import Any, Callable, Generic, Tuple

from approvaltests import verify_all
from approvaltests.core.options import Options
from approvaltests.inline.types import NT1, NT2, NT3, T1, T2, T3


class Parse3(Generic[T1, T2, T3]):
    def __init__(
        self,
        text: str,
        transformer: Callable[[str], Tuple[T1, T2, T3]],
        options: Options,
    ) -> None:
        self.text = text
        self._transformer = transformer
        self.options = options

    def verify_all(self, transform: Callable[[T1, T2, T3], Any]) -> None:
        from approvaltests.inline.parse import Parse

        verify_all(
            "",
            Parse.parse_inputs(self.text, self._transformer),
            lambda s: f"{s[0]}, {s[1]}, {s[2]} -> {transform(s[0], s[1], s[2])}",
            options=self.options.inline(),
        )

    def transform3(
        self,
        transform1: Callable[[T1], NT1],
        transform2: Callable[[T2], NT2],
        transform3: Callable[[T3], NT3],
    ) -> "Parse3[NT1, NT2, NT3]":
        def transformer(s: str) -> Tuple[NT1, NT2, NT3]:
            t1, t2, t3 = self._transformer(s)
            return (transform1(t1), transform2(t2), transform3(t3))

        return Parse3(self.text, transformer, self.options)



from pathlib import Path

from approvaltests.internals.logs.approved_file_log import ApprovedFilesLog
from approvaltests.internals.logs.log_commons import LogCommons


class FailedComparisonLog:
    @staticmethod
    def clear_log_file() -> None:
        FailedComparisonLog.get_failed_comparison_log().write_text("")
        LogCommons.download_script_from_common_repo_if_needed("approve_all.py")

    @staticmethod
    def get_failed_comparison_log() -> Path:
        return ApprovedFilesLog.get_temp_directory() / ".failed_comparison.log"

    @staticmethod
    def log(received_file: str, approved_file: str) -> None:
        with FailedComparisonLog.get_failed_comparison_log().open(mode="a") as file:
            file.write(f"{received_file} -> {approved_file}\n")


FailedComparisonLog.clear_log_file()



from typing import Optional

from approvaltests.namer.namer_base import NamerBase
from approvaltests.namer.stack_frame_namer import StackFrameNamer


def get_default_namer(extension: Optional[str] = None) -> NamerBase:
    return StackFrameNamer(extension)



import os
import subprocess
from typing import Any, Dict, Optional, Sequence

from approvaltests import Options, verify


def verify_command_line_with_inputs(
    command: str,
    *,
    inputs: Optional[Sequence[Any]] = None,
    options: Optional[Options] = None,
) -> None:
    input_string = "\n".join(map(lambda a: f"{a}", inputs))
    verify_command_line(command, input_string=input_string, options=options)


def verify_command_line(
    command_line: str,
    *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/
    input_string: Optional[str] = None,
    options: Optional[Options] = None,
    current_working_directory: str = ".",
    additional_environment_variables: Optional[Dict[str, str]] = None,
) -> None:
    # Set up environment with UTF-8 support for Windows
    my_env = {**os.environ}

    # Ensure UTF-8 encoding on Windows for emoji support
    if os.name == "nt":  # Windows
        my_env.update({"PYTHONIOENCODING": "utf-8", "PYTHONUTF8": "1"})

    if additional_environment_variables:
        my_env.update(additional_environment_variables)
    try:
        result = subprocess.run(
            command_line,
            check=False,
            shell=True,
            text=True,
            encoding="utf-8",
            input=input_string,
            cwd=current_working_directory,
            env=my_env,
            capture_output=True,
        )
        output = result.stdout
        if result.stderr:
            output += "\n--- STDERR ---\n" + result.stderr
    except UnicodeDecodeError:
        # Fallback to system default encoding if UTF-8 fails
        result = subprocess.run(
            command_line,
            check=False,
            shell=True,
            text=True,
            input=input_string,
            cwd=current_working_directory,
            env=my_env,
            capture_output=True,
        )
        output = result.stdout
        if result.stderr:
            output += "\n--- STDERR ---\n" + result.stderr
    verify(
        output,
        options=options,
    )



import sys
from pathlib import Path

_setup_dir = Path(__file__).resolve().parent / "setup"
if _setup_dir.is_dir():
    sys.path.insert(0, str(_setup_dir))
from setup_utils import do_the_setup, get_requirements_from_file

required = get_requirements_from_file("../requirements.prod.required.txt")

extra_requires = {
    "ClipboardReporter": ["pyperclip"],
    "verify_html": ["beautifulsoup4"],
    "PairwiseCombinations": ["allpairspy"],
}

do_the_setup(
    package_name="approvaltests-minimal",
    package_description="Assertion/verification library to aid testing with the minimal required dependencies + the ability to opt in for the others ",
    required=required,
    extra_requires=extra_requires,
)



from approvaltests.reporters.multi_reporter import MultiReporter
from approvaltests.reporters.python_native_reporter import PythonNativeReporter
from approvaltests.reporters.report_by_creating_diff_file import (
    ReportByCreatingDiffFile,
)


class ReportOnCyberDojo(MultiReporter):
    def __init__(self) -> None:
        super().__init__(ReportByCreatingDiffFile(), PythonNativeReporter())



from approvaltests.reporters.generic_diff_reporter_factory import (
    GenericDiffReporterFactory,
)

from .diff_tools import (
    ReportWithDiffToolOnLinux,
    ReportWithDiffToolOnMac,
    ReportWithDiffToolOnWindows,
)
from .first_working_reporter import FirstWorkingReporter
from .python_native_reporter import PythonNativeReporter


class DiffReporter(FirstWorkingReporter):
    """
    The DiffReporter class goes through a chain of possible diffing tools,
    to find the first option installed on your system.

    If none are found, it falls back to writing the diffs on
    the console.

    At present, the default Reporter is the DiffReporter.
    """

    def __init__(self, reporter_factory=None) -> None:
        factory = reporter_factory or GenericDiffReporterFactory()

        reporters = list(factory.get_all_reporters_from_config())
        reporters.extend(
            [
                ReportWithDiffToolOnWindows(),
                ReportWithDiffToolOnMac(),
                ReportWithDiffToolOnLinux(),
                PythonNativeReporter(),
            ]
        )
        super().__init__(*reporters)



import pathlib
from typing import Optional

from typing_extensions import override

from approval_utilities.approvaltests.core.executable_command import ExecutableCommand
from approval_utilities.utilities.multiline_string_utils import remove_indentation_from
from approvaltests.core.reporter import Reporter


class ExecutableCommandReporter(Reporter):
    def __init__(self, executor: ExecutableCommand, reporter: Reporter):
        self.executor = executor
        self.reporter = reporter

    @override
    def report(self, received_filename: str, approved_filename: str) -> bool:
        # recieved and approved commands are not the same
        # todo run the content of the file against the executable command
        # execute(read(received))

        self.reporter.report(
            self.execute_result(received_filename),
            self.execute_result(approved_filename),
        )
        self.reporter.report(received_filename, approved_filename)
        return True

    def execute_result(self, filename: str) -> str:
        path = pathlib.Path(filename)
        command_string = None
        if path.exists():
            command_string = path.read_text()
        result = ExecutableCommandReporter.execute_command_and_format_result(
            command_string, self.executor
        )
        approved_executed_result_file = (
            f"{path.name[: -len(path.suffix)]}.executed_results.txt"
        )
        pathlib.Path(approved_executed_result_file).write_text(result)
        return approved_executed_result_file

    @staticmethod
    def execute_command_and_format_result(
        my_command: Optional[str], executor: ExecutableCommand
    ) -> str:
        if not my_command:
            return ""

        result = executor.execute_command(my_command)
        return (
            remove_indentation_from(
                f"""
                Do NOT approve
                This File will be Deleted
                it is for feedback purposes only
        
        command: {my_command}
        
        result:
        """
            )
            + result
        )



import os
from typing import Optional


class Command:
    def __init__(self, cmd: str) -> None:
        self.command = cmd

    @staticmethod
    def executable(cmd: str) -> bool:
        return os.path.isfile(cmd) and os.access(cmd, os.X_OK)

    def locate(self) -> Optional[str]:
        path, _ = os.path.split(self.command)
        if path and self.executable(self.command):
            return self.command

        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe = os.path.join(path, self.command)
            if self.executable(exe):
                return exe

        return None



from typing import Callable, Dict, Optional

from approvaltests.core.comparator import Comparator
from approvaltests.core.namer import Namer
from approvaltests.core.reporter import Reporter
from approvaltests.file_approver import FileComparator
from approvaltests.inline.inline_options import InlineOptions
from approvaltests.scrubbers import combine_scrubbers


class FileOptions:
    def __init__(self, fields: Dict):
        self.fields = fields

    @property
    def file_extention(self) -> str:
        return self.fields.get("extension_with_dot", ".txt")

    def with_extension(
        self,
        extension_with_dot: str,
        *,  # enforce keyword arguments - https://www.python.org/dev/peps/pep-3102/,
        no_override: bool = False,
    ) -> "Options":
        if not extension_with_dot.startswith("."):
            extension_with_dot = "." + extension_with_dot
        if no_override and "extension_with_dot" in self.fields:
            extension_with_dot = self.fields["extension_with_dot"]
        return Options({**self.fields, "extension_with_dot": extension_with_dot})


class Options:
    def __init__(self, fields: Optional[Dict] = None):
        self.fields = fields or {}

    @property
    def reporter(self) -> Reporter:
        from approvaltests.reporters.default_reporter_factory import (
            get_default_reporter,
        )

        return self.fields.get("reporter") or get_default_reporter()

    @property
    def comparator(self) -> Comparator:
        return self.fields.get("comparator") or FileComparator()

    def with_comparator(self, comparator: Comparator) -> "Options":
        return Options({**self.fields, "comparator": comparator})

    def scrub(self, data: str) -> str:
        if self.has_scrubber():
            return self.fields["scrubber_func"](data)
        return data

    def with_scrubber(self, scrubber_func: Callable[[str], str]) -> "Options":
        return Options({**self.fields, "scrubber_func": scrubber_func})

    def add_scrubber(self, scrubber: Callable[[str], str]) -> "Options":
        if self.has_scrubber():
            scrubber = combine_scrubbers(self.fields["scrubber_func"], scrubber)
        return self.with_scrubber(scrubber)

    def has_scrubber(self) -> bool:
        return "scrubber_func" in self.fields

    def with_reporter(self, reporter: "Reporter") -> "Options":
        return Options({**self.fields, "reporter": reporter})

    def with_namer(self, namer: "Namer") -> "Options":
        return Options({**self.fields, "namer": namer})

    @property
    def for_file(self) -> FileOptions:
        return FileOptions(self.fields)

    @property
    def namer(self) -> Namer:
        from approvaltests.namer.default_name import get_default_namer

        namer = self.fields.get("namer") or get_default_namer()
        if hasattr(namer, "set_extension"):
            namer.set_extension(self.for_file.file_extention)
        return namer

    def inline(self, inline_options: Optional[InlineOptions] = None) -> "Options":
        from approvaltests.namer.inline_comparator import InlineComparator

        return InlineComparator().register(self, inline_options)

    def add_reporter(self, additional_reporter: Reporter) -> "Options":
        from approvaltests.reporters import MultiReporter

        return self.with_reporter(MultiReporter(self.reporter, additional_reporter))



import datetime
import re
import threading
from contextlib import contextmanager
from typing import Callable, Generator

from approval_utilities.utilities.logger.simple_logger import SimpleLogger
from approval_utilities.utilities.time_utilities import use_utc_timezone
from approvaltests import (
    Options,
    approvals,
    run_all_combinations,
    verify,
    verify_logging_for_all_combinations,
)
from approvaltests.utilities.logger.simple_logger_approvals import verify_simple_logger


def test_warnings() -> None:
    def scrubber(text: str) -> str:
        return re.sub(
            re.escape(__file__), "test_simple_logger.py", text, flags=re.IGNORECASE
        )

    output = SimpleLogger.log_to_string()
    SimpleLogger._wrapper.get().log_stack_traces = True
    text = "EVERYTHING IS AWFUL!!!!!!"
    try:
        raise Exception("EVERYTHING IS exceptionally AWFUL!!!!!!")
    except Exception as e:
        exception = e
    SimpleLogger.warning(text)
    SimpleLogger.warning(exception)
    SimpleLogger.warning(text, exception)
    verify(output, options=Options().with_scrubber(scrubber))


def log_from_inner_method() -> None:
    with SimpleLogger.use_markers():
        name = "Example"
        SimpleLogger.variable("name", name)
        for _ in range(0, 142):
            SimpleLogger.hour_glass()


def test_standard_logger() -> None:
    with verify_simple_logger(), SimpleLogger.use_markers() as m:
        log_from_inner_method()


def test_timestamps() -> None:
    with use_utc_timezone(), verify_simple_logger():
        count = -1

        def create_applesauce_timer() -> datetime.datetime:
            dates = [
                datetime.datetime.fromtimestamp(0.0),
                datetime.datetime.fromtimestamp(0.5),
                datetime.datetime.fromtimestamp(2.0),
                datetime.datetime.fromtimestamp(1050),
                datetime.datetime.fromtimestamp(1052),
            ]
            nonlocal count
            count = count + 1
            return dates[count]

        SimpleLogger._wrapper.get().timer = create_applesauce_timer
        SimpleLogger.show_timestamps(True)
        SimpleLogger.event("1")
        SimpleLogger.event("2")
        SimpleLogger.event("3")
        SimpleLogger.event("4")
        SimpleLogger.warning(exception=Exception("Oh no you didn't!"))


# begin-snippet: verify_simple_logger_example
def test_variable() -> None:
    with verify_simple_logger():
        SimpleLogger.variable("dalmatians", 101, show_types=True)
        SimpleLogger.variable("dalmatians", 101, show_types=False)


# end-snippet


# begin-snippet: verify_simple_logger_long_example
def test_variable_explict() -> None:
    output = SimpleLogger.log_to_string()
    SimpleLogger.variable("dalmatians", 101, show_types=True)
    SimpleLogger.variable("dalmatians", 101, show_types=False)
    verify(output)


# end-snippet


def test_variable_with_list() -> None:
    output = SimpleLogger.log_to_string()
    with SimpleLogger.use_markers():
        names = ["Jacqueline", "Llewellyn"]
        SimpleLogger.variable("names", names, show_types=True)
        SimpleLogger.variable("names", names, show_types=False)
    verify(output)


def verify_toggle(toggle_name: str, toggle: Callable[[bool], None]) -> None:
    SimpleLogger.show_all(True)
    SimpleLogger.event(f"Toggle Off {toggle_name}")
    toggle(False)
    log_everything()


def test_switching() -> None:
    output = SimpleLogger.log_to_string()

    verify_toggle("None", lambda a: SimpleLogger.show_all(True))
    verify_toggle("All", SimpleLogger.show_all)
    verify_toggle("Query", SimpleLogger.show_queries)
    verify_toggle("Message", SimpleLogger.show_messages)
    verify_toggle("Variable", SimpleLogger.show_variables)
    verify_toggle("Hour Glass", SimpleLogger.show_hour_glass)
    verify_toggle("Markers", SimpleLogger.show_markers)
    verify_toggle("Events", SimpleLogger.show_events)

    verify(output)


def log_everything() -> None:
    with SimpleLogger.use_markers():
        SimpleLogger.query("Select * from people")
        SimpleLogger.variable("Nonsense", "foo")
        SimpleLogger.event("Testing")
        SimpleLogger.message("Something random")
        for a in range(1, 13):
            SimpleLogger.hour_glass()
        try:
            infinity = 1 / 0
        except Exception as e:
            SimpleLogger.warning(exception=e)


def function_to_run(color: str, number: str) -> None:
    with SimpleLogger.use_markers():
        SimpleLogger.variable("color", color)
        SimpleLogger.variable("number", number)
        if number == "brie":
            raise Exception("AHHHHHH!")


def test_use_markers_with_raised_exception() -> None:
    def throw_exception() -> None:
        with SimpleLogger.use_markers():
            raise Exception("Everything is awflu!!?")

    output = SimpleLogger.log_to_string()
    try:
        throw_exception()
    except BaseException as e:
        SimpleLogger.warning(e)
    verify(output)


def test_run_combinations() -> None:
    with verify_simple_logger():
        run_all_combinations(function_to_run, [["red", "blue"], ["one", "two", "brie"]])


def test_run_combinations_with_exception_handler() -> None:
    approvals.settings().allow_multiple_verify_calls_for_this_method()

    with verify_simple_logger():
        run_all_combinations(
            function_to_run,
            [["red"], ["one", "brie"]],
            exception_handler=lambda e: SimpleLogger.warning(exception=e),
        )
    with verify_simple_logger():
        run_all_combinations(
            function_to_run,
            [["red"], ["one", "brie"]],
            exception_handler=lambda e: SimpleLogger.warning(e),
        )
    with verify_simple_logger():
        run_all_combinations(
            function_to_run,
            [["red"], ["one", "brie"]],
            exception_handler=SimpleLogger.warning,
        )


def test_verify_logging_for_all_combinations() -> None:
    verify_logging_for_all_combinations(
        function_to_run, [["red", "blue"], ["one", "two", "brie"]]
    )


# begin-snippet: method_with_inputs
def method_with_inputs(number: int, name: str) -> None:
    with SimpleLogger.use_markers(f"number = {number}, name = {name}"):
        # end-snippet
        print(f"{number}) {name}")


def test_markers_with_signature() -> None:
    output = SimpleLogger.log_to_string()
    method_with_inputs(1, "Susan")
    verify(output)


# begin-snippet: method_with_inputs_and_outputs
def method_with_inputs_and_outputs(number: int, announcement: str) -> None:
    with SimpleLogger.use_markers(
        lambda: f"number = {number}, announcement = {announcement}"
    ):
        # end-snippet
        for number in range(number, 0, -1):
            print(number)
        print(announcement)


def test_markers_with_signature_in_and_out() -> None:
    output = SimpleLogger.log_to_string()
    method_with_inputs_and_outputs(10, "Blast off")
    verify(output)


def test_race_condition() -> None:
    whose_turn = 0

    @contextmanager
    def wait_for(number: int) -> Generator[None, None, None]:
        nonlocal whose_turn
        while whose_turn < number:
            pass
        yield
        whose_turn += 1

    log1 = "Log1"

    def thread_1() -> None:
        nonlocal log1
        with wait_for(0):
            log1 = SimpleLogger.log_to_string(True)
        with wait_for(2):
            SimpleLogger.event("event_a")
        with wait_for(4):
            SimpleLogger.event("event_b")

    log2 = "Log2"

    def thread_2() -> None:
        nonlocal log2
        with wait_for(1):
            log2 = SimpleLogger.log_to_string(True)
        with wait_for(3):
            SimpleLogger.event("event_a")
        with wait_for(5):
            SimpleLogger.event("event_b")

    threading.Thread(target=thread_1).start()
    threading.Thread(target=thread_2).start()

    with wait_for(6):
        assert f"'{log1}'" == f"'{log2}'"
